import asyncio
import enum
import cgi


class ASGICycleState(enum.Enum):
    REQUEST = enum.auto()
    RESPONSE = enum.auto()


class ASGICycle:
    def __init__(self, scope: dict, body: bytes = b"", **kwargs) -> None:
        """
        Base for implementing the ASGI application request-response cycle for a
        particular FaaS platform.
        """
        self.scope = scope
        self.body = body
        self.state = ASGICycleState.REQUEST
        self.app_queue = None
        self.response = {}
        self.charset = None
        self.mimetype = None

    def __call__(self, app) -> dict:
        """
        Run the event loop and instantiate the ASGI application for the current request.
        """
        loop = asyncio.new_event_loop()
        self.app_queue = asyncio.Queue(loop=loop)
        self.put_message({"type": "http.request", "body": self.body})
        asgi_instance = app(self.scope)
        asgi_task = loop.create_task(asgi_instance(self.receive, self.send))
        loop.run_until_complete(asgi_task)
        return self.response

    def put_message(self, message: dict) -> None:
        self.app_queue.put_nowait(message)

    async def receive(self) -> dict:
        """
        An awaitable used by the application to receive messages from the queue.
        """
        message = await self.app_queue.get()
        return message

    async def send(self, message: dict) -> None:
        """
        An awaitable used by the application to send messages to the handler.
        """
        message_type = message["type"]

        if self.state is ASGICycleState.REQUEST:
            if message_type != "http.response.start":
                raise RuntimeError(
                    f"Expected 'http.response.start', received: {message_type}"
                )

            status_code = message["status"]
            headers = {k: v for k, v in message.get("headers", [])}

            if b"content-type" in headers:
                mimetype, options = cgi.parse_header(
                    headers[b"content-type"].decode("utf-8")
                )
                charset = options.get("charset", None)
                if charset:
                    self.charset = charset
                if mimetype:
                    self.mimetype = mimetype

            self.on_response_start(headers, status_code)
            self.state = ASGICycleState.RESPONSE

        elif self.state is ASGICycleState.RESPONSE:
            if message_type != "http.response.body":
                raise RuntimeError(
                    f"Expected 'http.response.body', received: {message_type}"
                )

            body = message["body"]

            self.on_response_body(body)
            self.put_message({"type": "http.disconnect"})

    def on_response_start(self, headers: list, status_code: int) -> None:
        """
        Handles the `http.response.start` event and begins building the response.
        """
        raise NotImplementedError()  # pragma: no cover

    def on_response_body(self, body: str) -> None:
        """
        Handles the `http.response.body` event and completes the response.
        """
        raise NotImplementedError()  # pragma: no cover



import click
import uuid
import boto3
import os
from mangum.platforms.aws.helpers import (
    get_default_resource_name,
    get_default_region_name,
    get_log_events,
    get_config,
)
from mangum.platforms.aws.config import AWSConfig


class AWSGroup(click.Group):
    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [x for x in self.list_commands(ctx) if x.startswith(cmd_name)]
        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail("Too many matches: %s" % ", ".join(sorted(matches)))


@click.command(cls=AWSGroup)
def aws() -> None:
    pass


@aws.command()
def init() -> None:
    click.echo(f"Welcome to Mangum, serverless ASGI!")
    project_name = click.prompt("What is the name of the project?", type=str)
    click.echo(f"{project_name}, great!")
    project_dir_name = click.prompt(
        "What is the name of the directory that contains the project code?", type=str
    )
    config_dir = os.getcwd()
    project_dir = os.path.join(config_dir, project_dir_name)
    if not os.path.isdir(project_dir):
        raise click.ClickException(
            f"Directory not found at {project_dir}! "
            "(Hint: The project folder must be in current working directory.)"
        )
    description = click.prompt(
        "Enter a brief description of your project.",
        type=str,
        default="ASGI application",
    )
    resource_name = click.prompt(
        "What should be the prefix used for naming resources?",
        type=str,
        default=get_default_resource_name(project_name),
    )
    url_root = click.prompt("What should be the root URL path?", type=str, default="/")
    runtime_version = click.prompt(
        "What version of Python are you using?", type=str, default="3.7"
    )
    timeout = click.prompt(
        "What should the timeout be (in seconds, max=300)?", type=int, default=300
    )
    default_region_name = get_default_region_name()
    region_name = click.prompt(
        "What region should be used?", default=default_region_name
    )
    s3_bucket_name = click.prompt(
        "An S3 bucket is required. \n\nEnter the name of an existing bucket or "
        f"one will be generated.",
        type=str,
        default="",
    )
    generate_s3 = s3_bucket_name == ""
    if generate_s3:
        s3_bucket_name = f"{resource_name.lower()}-{uuid.uuid4()}"
        s3 = boto3.resource("s3")
        s3.create_bucket(
            Bucket=s3_bucket_name,
            CreateBucketConfiguration={"LocationConstraint": region_name},
        )

    config = AWSConfig(
        project_name=project_name,
        description=description,
        s3_bucket_name=s3_bucket_name,
        resource_name=resource_name,
        url_root=url_root,
        runtime_version=runtime_version,
        region_name=region_name,
        timeout=timeout,
        stack_name=resource_name.lower(),
    )
    click.echo("Creating your local project...")
    config.init()
    click.echo(
        "Your app configuration has been generated!\n"
        "Modify the 'template.yaml' or 'requirements.txt' if you need to, otherwise "
        "run 'mangum aws build' to prepare the local project for packaging."
    )


@aws.command()
def build() -> None:
    """
    Install the packages listed in the requirements file and copy the application files
    to the 'build/' directory to prepare for packaging.
    """
    config, error = get_config()
    if error is not None:
        click.echo(error)
    else:
        click.echo("Building application and installing requirements...")
        config.build()
        click.echo("Build complete!")


@aws.command()
def update() -> None:
    """
    Update the application files only.
    """
    config, error = get_config()
    if error is not None:
        click.echo(error)
    else:
        click.echo("Updating the application build.")
        config.build(update=True)
        click.echo("Update complete!")


@aws.command()
def package() -> None:
    config, error = get_config()
    if error is not None:
        click.echo(error)
    else:
        click.echo("Packaging your application...")
        packaged = config.cli_package()
        if not packaged:
            click.echo("There was an error...")
        else:
            click.echo(
                "Successfully packaged. Run 'mangum aws deploy' to deploy it now."
            )


@aws.command()
def deploy() -> None:
    config, error = get_config()
    if error is not None:
        click.echo(error)
    else:
        click.echo("Deploying your application! This may take some time...")
        deployed = config.cli_deploy()
        if not deployed:
            click.echo("There was an error...")
        else:
            endpoints = config.cli_describe()
            click.echo(
                f"Deployment successful! API endpoints available at:\n\n{endpoints}"
            )


@aws.command()
def describe() -> None:
    config, error = get_config()
    if error is not None:
        click.echo(error)
    else:
        endpoints = config.cli_describe()
        if not endpoints:
            click.echo("Error! Could not retrieve endpoints.")
        else:
            click.echo(f"API endpoints available at:\n\n{endpoints}")


@aws.command()
def validate() -> None:
    config, error = get_config()
    if error is not None:
        click.echo(error)
    else:
        error = config.validate()
        if not error:
            click.echo("Template file validated successfully!")
        else:
            click.echo(error)


@aws.command()
def tail() -> None:
    config, error = get_config()
    if error is not None:
        click.echo(error)
    else:
        # Display the CloudWatch logs for the last 10 minutes.
        # TODO: Make this configurable.
        log_events = get_log_events(
            f"/aws/lambda/{config.resource_name}Function", minutes=10
        )
        log_output = []
        for log in log_events:
            message = log["message"].rstrip()
            if not any(
                i in message
                for i in ("START RequestId", "REPORT RequestId", "END RequestId")
            ):
                timestamp = log["timestamp"]
                s = f"[{timestamp}] {message}"
                log_output.append(s)

        click.echo("\n".join(log_output))



from typing import Any


class ServerlessAdapter:
    def __init__(self, app, debug: bool = False) -> None:
        """
        Base class for implementing platform-specific ASGI Adapter.
        """
        self.app = app
        self.debug = debug

    def __call__(self, *args, **kwargs) -> Any:
        """
        Attempt to run the application request-response cycle for a particular platform
        implementation.

        If an unhandled error is raised within the application, then optionally allow
        returning the error in the response if a debug response method exists.
        """
        try:
            response = self.asgi(*args, **kwargs)
        except Exception as exc:
            if self.debug:
                return self._debug(str(exc))
            raise exc
        else:
            return response

    def asgi(self, *args, **kwargs) -> Any:
        """
        This should be used as the main entrypoint for HTTP request events received
        from a platform.

        It should handle building a connection scope from the request information and
        running the ASGI request-response cycle, finally returning a valid response to
        the platform.
        """
        raise NotImplementedError()

    def _debug(self, content: str, status_code: int = 500) -> Any:
        """
        This should be used to send debug responses to a platform in the event of an
        unhandled server error. The debug behaviour is disabled by default, it should
        NOT be used in production environments.
        """
        raise NotImplementedError()



import urllib.parse
import os
import json


def encode_query_string(query_string_params: dict) -> str:
    return urllib.parse.urlencode(query_string_params).encode()


def get_file_content(
    *, filename: str, directory: str, as_json: bool = False
) -> str:  # pragma: no cover
    if not os.path.isdir(directory):
        raise IOError(f"Directory not found: '{directory}' does not exist.")
    filepath = os.path.join(directory, filename)
    if not os.path.exists(filepath):
        raise IOError(f"File not found: '{filepath}' does not exist.")
    with open(filepath, "r") as f:
        content = f.read()
    if as_json:
        try:
            content = json.loads(json.loads(content))
        except json.decoder.JSONDecodeError:
            raise ValueError(f"Invalid JSON data: '{filepath}' could not be decoded.")
    return content


def write_file_content(
    *, content: str, filename: str, directory: str, as_json: bool = False
) -> None:  # pragma: no cover
    if not os.path.isdir(directory):
        raise IOError(f"Directory not found: '{directory}' does not exist.")
    filepath = os.path.join(directory, filename)
    if as_json:
        content = json.dumps(content)
    with open(filepath, "w") as f:
        f.write(content)



import os
import subprocess
import json
import sys
from typing import Union
import boto3
import shutil
from mangum.utils import write_file_content


class AWSConfig:
    def __init__(
        self,
        *,
        project_name: str,
        description: str,
        region_name: str,
        runtime_version: str,
        url_root: str,
        s3_bucket_name: str,
        resource_name: str,
        timeout: int,
        stack_name: str,
    ) -> None:
        self.project_name = project_name
        self.description = description
        self.region_name = region_name
        self.runtime_version = runtime_version
        self.url_root = url_root
        self.s3_bucket_name = s3_bucket_name
        self.resource_name = resource_name
        self.timeout = timeout
        self.stack_name = stack_name

    @property
    def config_dir(self):
        return os.getcwd()

    @property
    def project_dir(self):
        return os.path.join(self.config_dir, self.project_name)

    @property
    def build_dir(self):
        build_path = os.path.join(self.config_dir, "build")
        return build_path

    def init(self) -> None:  # pragma: no cover
        """
        Write the configuration files used by the CLI. These will be used later to
        inform the AWS CLI wrappers.
        """
        config_files = {
            "template.yaml": {
                "directory": self.config_dir,
                "content": self.get_SAM_template(),
            },
            "settings.json": {
                "directory": self.config_dir,
                "content": json.dumps(self.get_build_context()),
                "as_json": True,
            },
            "requirements.txt": {"directory": self.config_dir, "content": "mangum\n"},
        }
        for name, info in config_files.items():
            write_file_content(
                content=info["content"],
                filename=name,
                directory=info["directory"],
                as_json=info.get("as_json", False),
            )

        shutil.copyfile(
            os.path.join(
                os.path.dirname(os.path.abspath(__file__)), "templates", "asgi.py"
            ),
            os.path.join(os.path.join(self.config_dir), "asgi.py"),
        )

    def build(self, update: bool = False) -> None:  # pragma: no cover
        """
        Copy all the files in the project directory to the build directory, then install
        the requirements.
        """
        project_build_dir = os.path.join(
            self.build_dir, os.path.basename(self.project_dir)
        )

        if update:
            print(project_build_dir)
            if os.path.exists(project_build_dir):
                shutil.rmtree(project_build_dir)
                print("RM")
            shutil.copytree(self.project_dir, project_build_dir)
        else:

            if os.path.exists(self.build_dir):
                shutil.rmtree(self.build_dir)

            shutil.copytree(self.project_dir, project_build_dir)

            for config_file in ("template.yaml", "asgi.py"):
                shutil.copyfile(
                    os.path.join(self.config_dir, config_file),
                    os.path.join(self.build_dir, config_file),
                )

            if not os.path.exists(os.path.join(self.config_dir, "requirements.txt")):
                raise IOError(f"File not found: 'requirements.txt' does not exist.")

            install_cmd = [
                sys.executable,
                "-m",
                "pip",
                "install",
                "-r",
                "requirements.txt",
                "-t",
                os.path.join(".", self.build_dir),
            ]
            installed = subprocess.run(install_cmd, stdout=subprocess.PIPE)
            if installed.returncode != 0:
                raise RuntimeError("Build failed, could not install requirements.")

    def validate(self) -> Union[None, str]:  # pragma: no cover
        with open(os.path.join(self.config_dir, "template.yaml")) as f:
            template_body = f.read()
        client = boto3.client("cloudformation")
        try:
            client.validate_template(TemplateBody=template_body)
        except Exception as exc:
            return str(exc)

    def cli_describe(self) -> Union[str, None]:  # pragma: no cover
        cmd = [
            "aws",
            "cloudformation",
            "describe-stacks",
            "--stack-name",
            self.stack_name,
            "--query",
            "Stacks[].Outputs",
        ]
        res = subprocess.run(cmd, stdout=subprocess.PIPE)
        if res.returncode != 0:
            return None
        data = json.loads(res.stdout)
        for i in data:
            for j in i:
                if j["OutputKey"] == f"{self.resource_name}Api":
                    endpoint = j["OutputValue"]
        return f"{endpoint}Prod\n\n{endpoint}Stage"

    def cli_package(self) -> bool:  # pragma: no cover

        cmd = [
            "aws",
            "cloudformation",
            "package",
            "--template-file",
            "template.yaml",
            "--output-template-file",
            "packaged.yaml",
            "--s3-bucket",
            f"{self.s3_bucket_name}",
        ]
        res = subprocess.run(cmd, stdout=subprocess.PIPE)
        return res.returncode == 0

    def cli_deploy(self) -> bool:  # pragma: no cover
        cmd = [
            "aws",
            "cloudformation",
            "deploy",
            "--template-file",
            "packaged.yaml",
            "--stack-name",
            f"{self.stack_name}",
            "--capabilities",
            "CAPABILITY_IAM",
        ]
        res = subprocess.run(cmd, stdout=subprocess.PIPE)
        return res.returncode == 0

    def get_build_context(self) -> dict:
        return {
            "project_name": self.project_name,
            "description": self.description,
            "region_name": self.region_name,
            "runtime_version": self.runtime_version,
            "resource_name": self.resource_name,
            "url_root": self.url_root,
            "s3_bucket_name": self.s3_bucket_name,
            "timeout": self.timeout,
            "stack_name": self.stack_name,
        }

    def get_SAM_template(self) -> str:
        return f"""AWSTemplateFormatVersion: '2010-09-09'
Transform: AWS::Serverless-2016-10-31
Description: >
    {self.project_name}
    {self.description}
Globals:
    Function:
        Timeout: {self.timeout}
Resources:
    {self.resource_name}Function:
        Type: AWS::Serverless::Function
        Properties:
            FunctionName: {self.resource_name}Function
            CodeUri: ./build
            Handler: asgi.handler
            Runtime: python{self.runtime_version}
            # Environment:
            #     Variables:
            #         PARAM1: VALUE
            Events:
                {self.resource_name}:
                    Type: Api
                    Properties:
                        Path: {self.url_root}
                        Method: get
Outputs:
    {self.resource_name}Api:
      Description: "API Gateway endpoint URL for {self.resource_name} function"
      Value: !Sub "https://${{ServerlessRestApi}}.execute-api.${{AWS::Region}}.amazonaws.com{self.url_root}"
    {self.resource_name}Function:
      Description: "{self.resource_name} Lambda Function ARN"
      Value: !GetAtt {self.resource_name}Function.Arn
    {self.resource_name}FunctionIamRole:
      Description: "Implicit IAM Role created for {self.resource_name} function"
      Value: !GetAtt {self.resource_name}FunctionRole.Arn"""



import datetime
import operator
import boto3
from typing import Tuple, Union
import os
from mangum.platforms.aws.config import AWSConfig
from mangum.utils import get_file_content


def get_config() -> Tuple[Union[AWSConfig, None], Union[None, str]]:  # pragma: no cover
    current_dir = os.getcwd()
    try:
        settings = get_file_content(
            filename="settings.json", directory=current_dir, as_json=True
        )
    except Exception as exc:
        return None, f"[Error] {exc}"
    config = AWSConfig(**settings)
    return config, None


def get_default_region_name() -> str:  # pragma: no cover
    session = boto3.session.Session()
    return session.region_name


def get_default_resource_name(project_name: str) -> str:
    if "_" in project_name:
        name_parts = project_name.split("_")
        resource_name = "".join([s.title() for s in name_parts])
    else:
        resource_name = project_name.title()
    return resource_name


def get_log_events(group_name: str, minutes: int) -> list:  # pragma: no cover
    end_dt = datetime.datetime.now()
    start_dt = end_dt - datetime.timedelta(minutes=minutes)
    start_time = int(start_dt.timestamp()) * 1000
    end_time = int(end_dt.timestamp()) * 1000

    kwargs = {
        "logGroupName": group_name,
        "startTime": start_time,
        "endTime": end_time,
        "limit": 10000,
    }

    all_events = []
    client = boto3.client("logs")

    while True:
        res = client.filter_log_events(**kwargs)
        all_events += res["events"]

        try:
            kwargs["nextToken"] = res["nextToken"]
        except KeyError:
            break

    return sorted(all_events, key=operator.itemgetter("timestamp"), reverse=False)



import base64
from mangum.utils import encode_query_string
from mangum.asgi.protocol import ASGICycle
from mangum.asgi.adapter import ServerlessAdapter


class AWSLambdaASGICycle(ASGICycle):
    """
    An adapter that handles the HTTP request-response cycle for an ASGI application and
    builds a valid response to return to AWS Lambda & API Gateway.

    The response is a Python dictionary with the following structure:

        {
            "statusCode": int,
            "isBase64Encoded": bool,
            "headers": dict,
            "body": str,
        }
    """

    def __init__(self, *args, **kwargs) -> None:
        self.binary = kwargs.pop("binary", False)
        super().__init__(*args, **kwargs)

    def on_response_start(self, headers: dict, status_code: int) -> None:
        self.response["statusCode"] = status_code
        self.response["isBase64Encoded"] = self.binary
        self.response["headers"] = {k.decode(): v.decode() for k, v in headers.items()}

    def on_response_body(self, body: bytes) -> None:
        if self.binary:
            body = base64.b64encode(body)
        else:
            body = body.decode("ascii")
        self.response["body"] = body


class AWSLambdaAdapter(ServerlessAdapter):
    """
    A adapter that wraps an ASGI application and handles transforming an incoming
    AWS Lambda & API Gateway request into the ASGI connection scope.

    After building the connection scope, it runs the ASGI application cycle and returns
    the response.
    """

    def asgi(self, event: dict, context: dict) -> dict:
        server = None
        client = None
        method = event["httpMethod"]
        headers = event["headers"] or {}
        path = event["path"]
        host = headers.get("Host")
        scheme = headers.get("X-Forwarded-Proto", "http")
        x_forwarded_for = headers.get("X-Forwarded-For")
        x_forwarded_port = headers.get("X-Forwarded-Port")
        if x_forwarded_port and x_forwarded_for:
            port = int(x_forwarded_port)
            client = (x_forwarded_for, port)
            if host:
                server = (host, port)
        query_string_params = event["queryStringParameters"]
        query_string = (
            encode_query_string(query_string_params) if query_string_params else b""
        )

        scope = {
            "server": server,
            "client": client,
            "scheme": scheme,
            "root_path": "",
            "query_string": query_string,
            "headers": [[k.encode(), v.encode()] for k, v in headers.items()],
            "type": "http",
            "http_version": "1.1",
            "method": method,
            "path": path,
        }

        binary = event.get("isBase64Encoded", False)
        body = event["body"]
        if body:
            if binary:
                body = base64.b64decode(body)
            elif not isinstance(body, bytes):
                body = body.encode("utf-8")

        response = AWSLambdaASGICycle(scope, body=body, binary=binary)(self.app)
        return response

    def _debug(self, content: str, status_code: int = 500) -> None:
        return {
            "statusCode": status_code,
            "isBase64Encoded": False,
            "headers": {},
            "body": content,
        }



import urllib.parse
from azure.functions import HttpRequest, HttpResponse
from mangum.asgi.protocol import ASGICycle
from mangum.asgi.adapter import ServerlessAdapter
from mangum.utils import encode_query_string


class AzureFunctionCycle(ASGICycle):
    """
    An adapter that handles the HTTP request-response cycle for an ASGI application and
    builds a valid response to return to Azure Functions.

    The response is a Python dictionary with the following structure:

        {
            "body": bytes,
            "charset": str,
            "headers": dict,
            "mimetype": str,
            "status_code": itnt
        }
    """

    def on_response_start(self, headers: dict, status_code: int) -> None:
        self.response["status_code"] = status_code
        self.response["headers"] = {k.decode(): v.decode() for k, v in headers.items()}
        self.response["mimetype"] = self.mimetype
        self.response["charset"] = self.charset

    def on_response_body(self, body: bytes) -> None:
        self.response["body"] = body


class AzureFunctionAdapter(ServerlessAdapter):
    """
    A adapter that wraps an ASGI application and handles transforming an incoming
    Azure Function request into the ASGI connection scope.

    After building the connection scope, it runs the ASGI application cycle and then
    serializes the response into an `HttpResponse`.
    """

    def asgi(self, event: HttpRequest) -> dict:
        server = None
        client = None
        scheme = "https"
        method = event.method
        headers = event.headers.items()
        parsed = urllib.parse.urlparse(event.url)
        scheme = parsed.scheme
        path = parsed.path
        query_string = encode_query_string(event.params) if event.params else b""

        scope = {
            "type": "http",
            "server": server,
            "client": client,
            "method": method,
            "path": path,
            "scheme": scheme,
            "http_version": "1.1",
            "root_path": "",
            "query_string": query_string,
            "headers": [[k.encode(), v.encode()] for k, v in headers],
        }

        body = event.get_body() or b""

        response = AzureFunctionCycle(scope, body=body)(self.app)
        return HttpResponse(
            body=response["body"],
            headers=response["headers"],
            status_code=response["status_code"],
            mimetype=response["mimetype"],
            charset=response["charset"],
        )



from setuptools import find_packages, setup

from mangum.__version__ import __version__


def get_long_description():
    return open("README.md", "r", encoding="utf8").read()


setup(
    name="mangum",
    version=__version__,
    packages=find_packages(),
    license="MIT",
    url="https://github.com/erm/mangum",
    description="ASGI serverless adapters",
    long_description=get_long_description(),
    extras_require={
        "aws": ["boto3", "click", "jinja2"],
        "azure": ["azure_functions"],
        "full": ["azure-functions", "boto3", "click", "jinja2"],
    },
    long_description_content_type="text/markdown",
    entry_points={"console_scripts": ["mangum = mangum.__main__:main"]},
    author="Jordan Eremieff",
    author_email="jordan@eremieff.com",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
)
